/*
 *  Copyright (C) 2003, 2004, 2005 Max-Planck-Society
 *  Author: Martin Reinecke
 */

#include "alm.h"
#include "healpix_map.h"
#include "fftpack_support.h"
#include "ylmgen.h"
#include "xcomplex.h"

using namespace std;

namespace {

void init_lam_fact_1d (int m, arr<double> &lam_fact)
  {
  for (int l=m; l<lam_fact.size(); ++l)
    lam_fact[l] = (l<2) ? 0. : 2*sqrt((2*l+1.)/(2*l-1.) * (l*l-m*m));
  }

void init_lam_fact_deriv_1d (int m, arr<double> &lam_fact)
  {
  lam_fact[m]=0;
  for (int l=m+1; l<lam_fact.size(); ++l)
    lam_fact[l] = sqrt((2*l+1.)/(2*l-1.) * (l*l-m*m));
  }

void init_normal_l (arr<double> &normal_l)
  {
  for (int l=0; l<normal_l.size(); ++l)
    normal_l[l] = (l<2) ? 0. : sqrt(1./((l+2.)*(l+1.)*l*(l-1.)));
  }

void get_chunk_info (int nrings, int &nchunks, int &chunksize)
  {
  nchunks = nrings/max(100,nrings/10) + 1;
  chunksize = (nrings+nchunks-1)/nchunks;
  }

} // namespace
#define SETUP_LAMBDA \
  const double t1  = lam_lm1m*lam_fact[l]; \
  const double a_w = (l-m2)*two_on_s2 + l*(l-1); \
  const double a_x = twocth*(l-1)*lam_lm; \
  const double lambda_w = a_w*lam_lm - t1*c_on_s2; \
  const double lambda_x = m_on_s2 * (a_x-t1);
#define ALM2MAP_POL_MACRO(bT1,bQ1,bQ2,bU1,bU2) \
  { \
  double lam_lm1m = lam_lm; \
  lam_lm = Ylm[l]; \
  SETUP_LAMBDA \
  bT1.re+=alm_tmp[l][0].re*lam_lm;   bT1.im+=alm_tmp[l][0].im*lam_lm; \
  bQ1.re+=alm_tmp[l][1].re*lambda_w; bQ1.im+=alm_tmp[l][1].im*lambda_w; \
  bQ2.re-=alm_tmp[l][2].im*lambda_x; bQ2.im+=alm_tmp[l][2].re*lambda_x; \
  bU1.re-=alm_tmp[l][2].re*lambda_w; bU1.im-=alm_tmp[l][2].im*lambda_w; \
  bU2.re-=alm_tmp[l][1].im*lambda_x; bU2.im+=alm_tmp[l][1].re*lambda_x; \
  }

template<typename T> void alm2map_pol
  (const Alm<xcomplex<T> > &almT,
   const Alm<xcomplex<T> > &almG,
   const Alm<xcomplex<T> > &almC,
   Healpix_Map<T> &mapT,
   Healpix_Map<T> &mapQ,
   Healpix_Map<T> &mapU)
  {
  planck_assert (mapT.Scheme()==RING,
    "alm2map_pol: maps must be in RING scheme");
  planck_assert (mapT.conformable(mapQ) && mapT.conformable(mapU),
    "alm2map_pol: maps are not conformable");
  planck_assert (almT.conformable(almG) && almT.conformable(almC),
    "alm2map_pol: a_lms are not conformable");

  int lmax = almT.Lmax(), mmax = almT.Mmax(), nside = mapT.Nside();

  arr<double> normal_l (lmax+1);
  init_normal_l (normal_l);

  int nchunks, chunksize;
  get_chunk_info(2*nside,nchunks,chunksize);

  arr2<xcomplex<double> >
    b_north_T(chunksize,mmax+1), b_south_T(chunksize,mmax+1),
    b_north_Q(chunksize,mmax+1), b_south_Q(chunksize,mmax+1),
    b_north_U(chunksize,mmax+1), b_south_U(chunksize,mmax+1);

  arr<double> cth(chunksize),sth(chunksize);

  for (int chunk=0; chunk<nchunks; ++chunk)
    {
    int llim=chunk*chunksize, ulim=min(llim+chunksize,2*nside);
    for (int ith=llim; ith<ulim; ++ith)
      {
      int nph, istart_north;
      bool shifted;
      mapT.get_ring_info (ith+1,istart_north,nph, cth[ith-llim],sth[ith-llim],
                          shifted);
      }

#pragma omp parallel
{
    Ylmgen generator(lmax,mmax,1e-30);
    arr<double> Ylm;
    arr<double> lam_fact (lmax+1);
    arr<xcomplex<double>[3]> alm_tmp(lmax+1);
    int m;
#pragma omp for schedule(dynamic,1)
    for (m=0; m<=mmax; ++m)
      {
      int m2 = m*m;
      init_lam_fact_1d (m,lam_fact);
      for (int l=m; l<=lmax; ++l)
        {
        alm_tmp[l][0] = almT(l,m);
        alm_tmp[l][1] = almG(l,m)*(-normal_l[l]);
        alm_tmp[l][2] = almC(l,m)*(-normal_l[l]);
        }
      for (int ith=0; ith<ulim-llim; ++ith)
        {
        int l;
        generator.get_Ylm(cth[ith],sth[ith],m,Ylm,l);
        if (l<=lmax)
          {
          double one_on_s2 = 1/(sth[ith]*sth[ith]);
          double c_on_s2 = cth[ith] * one_on_s2;
          double two_on_s2 = 2*one_on_s2;
          double m_on_s2 = m*one_on_s2;
          double twocth = 2*cth[ith];

          xcomplex<double> bT1=0, bT2=0, bQ1=0, bQ2=0, bU1=0, bU2=0;

          double lam_lm = 0;
          if ((l-m)&1) goto middle;
start:    ALM2MAP_POL_MACRO(bT1,bQ1,bQ2,bU1,bU2)
          if (++l>lmax) goto end;
middle:   ALM2MAP_POL_MACRO(bT2,bQ2,bQ1,bU2,bU1)
          if (++l<=lmax) goto start;
end:      b_north_T[ith][m] = bT1+bT2; b_south_T[ith][m] = bT1-bT2;
          b_north_Q[ith][m] =-bQ1-bQ2; b_south_Q[ith][m] =-bQ1+bQ2;
          b_north_U[ith][m] = bU1+bU2; b_south_U[ith][m] = bU1-bU2;
          }
        else
          {
          b_north_T[ith][m] = b_south_T[ith][m] = 0;
          b_north_Q[ith][m] = b_south_Q[ith][m] = 0;
          b_north_U[ith][m] = b_south_U[ith][m] = 0;
          }
        }
      }
} // end of parallel region

    }
  }

template void alm2map_pol (const Alm<xcomplex<double> > &almT,
                           const Alm<xcomplex<double> > &almG,
                           const Alm<xcomplex<double> > &almC,
                           Healpix_Map<double> &mapT,
                           Healpix_Map<double> &mapQ,
                           Healpix_Map<double> &mapU);

template void alm2map_pol (const Alm<xcomplex<float> > &almT,
                           const Alm<xcomplex<float> > &almG,
                           const Alm<xcomplex<float> > &almC,
                           Healpix_Map<float> &mapT,
                           Healpix_Map<float> &mapQ,
                           Healpix_Map<float> &mapU);
